part of nueli.core.bucket;

Palette get palette => Palette.configure();

class Palette extends Bucket<Color> {
  Palette.configure() : super.configure();

  static Symbol _themeChannel = #basic;
  static set theme(Symbol themeCode) => _themeChannel = themeCode;

  Symbol get _iD => #palette;
}

class Swatch extends Color {
  Swatch(int basic, [Map<Symbol, int> others = const {}])
      : super(handle(
              () => {#basic: basic, ...others}[Palette._themeChannel],
              alt: (e) => {#basic: basic, ...others}[#basic],
            ) ??
            basic);
}

extension Palette_Nueli on Palette {
  Color get nueliGrey => Swatch(0xff45494e, {#dark: 0xffffffff});
  Color get nueliOrange => Swatch(0xffe77d23, {#dark: 0xfffdb880});
}

extension Palette_Material on Palette {
  static void configure({
    MaterialColor? material,
    MaterialAccentColor? materialAccent,
  }) {
    (palette * #material)[Colors.blueGrey] = material;
    (palette * #materialAccent)[Colors.orangeAccent] = materialAccent;
  }

  Color get material => (palette * #material)[Colors.blue];
  Color get materialAccent => (palette * #materialAccent)[Colors.tealAccent];
}

extension Palette_Surface on Palette {
  static void configure({
    Color? primary,
    Color? primaryLight,
    Color? primaryDark,
    Color? accent,
    Color? background,
  }) {
    (palette * #primary)[(palette * #material)[Colors.blueGrey]] = primary;
    (palette * #primaryLight)[(palette * #material)[Colors.blueGrey]] = primaryLight;
    (palette * #primaryDark)[(palette * #material)[Colors.blueGrey]] = primaryDark;
    (palette * #accent)[(palette * #materialAccent)[Colors.blueGrey]] = accent;
    (palette * #background)[Colors.white] = background;
  }

  Color get primary => (palette * #primary)[Colors.blueGrey];
  Color get primaryLight => (palette * #primaryLight)[Colors.blueGrey];
  Color get primaryDark => (palette * #primaryDark)[Colors.blueGrey];
  Color get accent => (palette * #accent)[Colors.blueGrey];
  Color get background => (palette * #background)[Colors.white];
}

extension Palette_Content on Palette {
  static void configure({
    Color? onPrimary,
    Color? onPrimaryLight,
    Color? onPrimaryDark,
    Color? onAccent,
    Color? onBackground,
  }) {
    (palette * #onBackground)[Colors.blueGrey] = onBackground;
    (palette * #onPrimary)[(palette * #background)[Colors.blueGrey]] = onPrimary;
    (palette * #onPrimaryLight)[(palette * #onBackground)[Colors.blueGrey]] = onPrimaryLight;
    (palette * #onPrimaryDark)[(palette * #background)[Colors.blueGrey]] = onPrimaryDark;
    (palette * #onAccent)[(palette * #onBackground)[Colors.black54]] = onAccent;
  }

  Color get onBackground => (palette * #onBackground)[Colors.blueGrey];
  Color get onPrimary => (palette * #onPrimary)[Colors.blueGrey];
  Color get onPrimaryLight => (palette * #onPrimaryLight)[Colors.blueGrey];
  Color get onPrimaryDark => (palette * #onPrimaryDark)[Colors.blueGrey];
  Color get onAccent => (palette * #onAccent)[Colors.blueGrey];
}

extension Palette_Color on Palette {
  static void configure({Color? shade, Color? tint}) {
    (palette * #shade)[Colors.black54] = shade;
    (palette * #tint)[Colors.white54] = tint;
  }

  Color get none => Colors.white.withAlpha(0);
  Color get white => Colors.white;
  Color get black => Colors.black;
  Color get grey => Colors.grey;
  Color get blueGrey => Colors.blueGrey;
  Color get shade => (palette * #shade)[Colors.black54];
  Color get tint => (palette * #tint)[Colors.white54];
}

extension Palette_Status on Palette {
  static void scheme({Color? neutral, Color? error, Color? success, Color? warning, Color? message}) {
    (palette * #neutral)[(palette * #accent)[Colors.blueGrey]] = neutral;
    (palette * #error)[Colors.redAccent] = error;
    (palette * #success)[Colors.green] = success;
    (palette * #warning)[Colors.amber] = warning;
    (palette * #message)[Colors.brown] = message;
  }

  Color get neutral => (palette * #neutral)[Colors.blueGrey];
  Color get error => (palette * #error)[Colors.redAccent];
  Color get success => (palette * #success)[Colors.green];
  Color get warning => (palette * #warning)[Colors.amber];
  Color get message => (palette * #message)[Colors.brown];
}
