part of nueli.core.config;

/// Ace shortcut
Ace get $ace => Ace();

/// Ace ~ [BoxShadow] configuration class.
class Ace extends Config<Ace, AceData> {
  List<Ace> _children = const [];
  Color? _color;
  Offset? _offset;
  double? _spread;
  double? _blur;

  Ace() : super(AceData());

  Ace _config({
    Color? color,
    Offset? offset,
    double? spread,
    double? blur,
  }) {
    _color = color ?? _color;
    _offset = offset ?? _offset;
    _spread = spread ?? _spread;
    _blur = blur ?? _blur;
    return this;
  }

  // variant commands
  Ace color(Color color, [int? alpha]) => _config(color: color.withAlpha(alpha ?? palette.shade.alpha));
  Ace offset(double x, [double? y]) => _config(offset: Offset(x, y ?? x));
  Ace blur(double blur) => _config(blur: blur);
  Ace spread(double spread) => _config(spread: spread);

  // fixed commands
  Ace withPosition(DecorationPosition? position) => this & position;

  // templates
  factory Ace.elevated(double elevation) => Ace()._config(blur: elevation, spread: -0.27, offset: Offset(elevation / 8, elevation / 8));
  factory Ace.raised(double float) => Ace()._config(blur: float / 6, spread: -0.5, offset: Offset(float, float));
  factory Ace.glow(double glow) => Ace()._config(blur: glow, spread: -0.1, offset: Offset.zero);

  @override
  Ace operator *(double factor) {
    final aceAp = _AceApprox(this);
    _children.map((e) => e * factor).toList();
    return _config(
      color: aceAp.color * factor,
      offset: aceAp.offset * factor,
      spread: aceAp.spread * factor,
      blur: aceAp.blur * factor,
    );
  }

  @override
  Ace operator -(Ace operand) {
    final aceAp = _AceApprox(this);
    final aceAp2 = _AceApprox(operand);
    _children.map((e) => e - operand).toList();
    return _config(
      color: aceAp.color - aceAp2.color,
      offset: aceAp.offset - aceAp2.offset,
      spread: aceAp.spread - aceAp2.spread,
      blur: aceAp.blur - aceAp2.blur,
    );
  }

  @override
  Ace operator +(Ace operand) {
    final aceAp = _AceApprox(this);
    final aceAp2 = _AceApprox(operand);
    _children.map((e) => e + operand).toList();
    return _config(
      color: aceAp.color + aceAp2.color,
      offset: aceAp.offset + aceAp2.offset,
      spread: aceAp.spread + aceAp2.spread,
      blur: aceAp.blur + aceAp2.blur,
    );
  }

  @override
  Ace lerp(shadow) {
    if (shadow is Ace)
      _config(
        color: shadow._color ?? _color,
        blur: shadow._blur ?? _blur,
        spread: shadow._spread ?? _spread,
        offset: shadow._offset ?? _offset,
      );
    return super.lerp(shadow);
  }

  BoxShadow get toBoxShadow {
    final aceAp = _AceApprox(this);
    return BoxShadow(
      color: aceAp.color,
      offset: aceAp.offset,
      blurRadius: aceAp.blur,
      spreadRadius: aceAp.spread,
    );
  }

  List<BoxShadow> get toBoxShadowDeep => _children.map((e) => e.toBoxShadow).toList()..add(this.toBoxShadow);

  Ace save() => Ace().._children.addAll([...this._children, this.._children = []]);

  Widget call() {
    return DecoratedBox(
      decoration: BoxDecoration(boxShadow: toBoxShadowDeep),
      child: data.child,
      position: data.position ?? DecorationPosition.background,
    );
  }
}

class AceData extends ConfigData<AceData> {
  DecorationPosition? position;

  AceData({Widget? child, this.position}) : super(child);

  AceData lerp(data) {
    if (data is DecorationPosition) position = data;

    if (data is AceData) position = data.position;
    return super.lerp(data);
  }
}

class AceTween extends Tween<Ace> {
  AceTween({Ace? begin, Ace? end}) : super(begin: begin, end: end);

  @override
  Ace transform(double value) => (begin ?? $ace) + ((end ?? $ace) - (begin ?? $ace)) * value;
}

class AceTransition extends AnimatedWidget {
  final Widget? child;
  AceTransition({required Animation<Ace> ace, this.child}) : super(listenable: ace);
  Animation<Ace> get ace => listenable as Animation<Ace>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => ace.value.withChild(child)(),
    );
  }
}

class _AceApprox {
  final Ace shadow;

  _AceApprox(this.shadow);
  Color get color => shadow._color ?? palette.none;
  Offset get offset => shadow._offset ?? Offset.zero;
  double get spread => shadow._spread ?? 0;
  double get blur => shadow._blur ?? 0;
}
