part of nueli.core.config;

/// Decor shortcut
Decor get $decor => Decor();

/// Decor ~ [BoxDecoration] configuration class.
class Decor extends Config<Decor, DecorData> {
  List<DecorPaint>? _paints;
  Alignment? _origin, _focus;
  Color? _stroke;
  double? _strokeSize;
  Radii? _radii;
  Ace? _shadow;
  Ink? _ink;

  Decor() : super(DecorData());

  Decor _config({
    List<DecorPaint>? paints,
    Alignment? center,
    Alignment? focus,
    Color? stroke,
    double? strokeSize,
    Radii? radii,
    Ace? shadow,
    Ink? ink,
  }) {
    _paints = paints ?? _paints;
    _origin = center ?? _origin;
    _focus = focus ?? _focus;
    _stroke = stroke ?? _stroke;
    _strokeSize = strokeSize ?? _strokeSize;
    _radii = radii ?? _radii;
    _shadow = shadow ?? _shadow;
    _ink = ink ?? _ink;
    return this;
  }

  // variant commands
  Decor fill(List<DecorPaint> hue) => _config(paints: hue);
  Decor paint(Color color, [double stop = 1]) => fill((_paints ?? [])..add(DecorPaint(color.value, stop)));
  Decor color(Color color) => _config(paints: [DecorPaint(color.value)]);
  Decor stroke(double thickness, [Color? color]) => _config(stroke: color ?? palette.accent, strokeSize: thickness);
  Decor orient(Alignment center, Alignment focus) => _config(center: center, focus: focus);

  // fixed commands
  Decor withSize(double width, [double? height]) => this & Size(width, height ?? width);
  Decor withPadding(EdgeInsets? padding) => this & padding;
  Decor withMargin(EdgeInsets? margin) => this & Attr(#margin)[margin];
  Decor withAlignment(Alignment? alignment) => this & alignment;
  Decor withConstraints(BoxConstraints? constraints) => this & constraints;
  Decor withStyle(DecorStyle? style) => this & style;
  Decor withClip(Clip? clip) => this & clip;

  // config commands
  Decor $radii(Radii radii) => _config(radii: radii);
  Decor $ace(Ace shadow) => _config(shadow: shadow);
  Decor $ink(Ink react) => _config(ink: react);

  // shortcuts
  Decor get paint_none => paint(palette.none); // ignore: non_constant_identifier_names
  Decor get paint_primary => paint(palette.primary); // ignore: non_constant_identifier_names
  Decor get paint_accent => paint(palette.accent); // ignore: non_constant_identifier_names
  Decor get paint_background => paint(palette.background); // ignore: non_constant_identifier_names
  Decor get paint_onBackground => paint(palette.onBackground); // ignore: non_constant_identifier_names

  Decor get color_none => color(palette.none); // ignore: non_constant_identifier_names
  Decor get color_primary => color(palette.primary); // ignore: non_constant_identifier_names
  Decor get color_accent => color(palette.accent); // ignore: non_constant_identifier_names
  Decor get color_background => color(palette.background); // ignore: non_constant_identifier_names
  Decor get color_onBackground => color(palette.onBackground); // ignore: non_constant_identifier_names

  Decor get stroke_none => stroke(0.5, palette.none); // ignore: non_constant_identifier_names
  Decor get stroke_primary => stroke(0.5, palette.primary); // ignore: non_constant_identifier_names
  Decor get stroke_accent => stroke(0.5, palette.accent); // ignore: non_constant_identifier_names
  Decor get stroke_background => stroke(0.5, palette.background); // ignore: non_constant_identifier_names
  Decor get stroke_onBackground => stroke(0.5, palette.onBackground); // ignore: non_constant_identifier_names

  Decor get orient_vert => orient(Alignment.topCenter, Alignment.bottomCenter); // ignore: non_constant_identifier_names
  Decor get orient_horz => orient(Alignment.centerLeft, Alignment.centerRight); // ignore: non_constant_identifier_names
  Decor get orient_diagonal => orient(Alignment.topLeft, Alignment.bottomRight); // ignore: non_constant_identifier_names

  @override
  Decor operator *(double factor) {
    final decorAp = _DecorApprox(this);
    return _config(
      paints: decorAp.paints.map((e) => e * factor).toList(),
      center: decorAp.origin * factor,
      shadow: decorAp.shadow * factor,
      stroke: decorAp.stroke * factor,
      radii: decorAp.radii * factor,
      strokeSize: decorAp.strokeSize * factor,
      ink: decorAp.ink * factor,
    );
  }

  @override
  Decor operator -(Decor operand) {
    final decorAp = _DecorApprox(this);
    final decorAp2 = _DecorApprox(operand);
    return _config(
      paints: decorAp.paints.minus(decorAp2.paints).toList(),
      center: decorAp.origin - decorAp2.origin,
      shadow: decorAp.shadow - decorAp2.shadow,
      stroke: decorAp.stroke - decorAp2.stroke,
      radii: decorAp.radii - decorAp2.radii,
      strokeSize: decorAp.strokeSize - decorAp2.strokeSize,
      ink: decorAp.ink - decorAp2.ink,
    );
  }

  @override
  Decor operator +(Decor operand) {
    final decorAp = _DecorApprox(this);
    final decorAp2 = _DecorApprox(operand);
    return _config(
      paints: decorAp.paints.minus(decorAp2.paints).toList(),
      center: decorAp.origin + decorAp2.origin,
      shadow: decorAp.shadow + decorAp2.shadow,
      stroke: decorAp.stroke + decorAp2.stroke,
      radii: decorAp.radii + decorAp2.radii,
      strokeSize: decorAp.strokeSize + decorAp2.strokeSize,
      ink: decorAp.ink + decorAp2.ink,
    );
  }

  @override
  Decor lerp(decor) {
    if (decor is Decor)
      _config(
        paints: decor._paints ?? _paints,
        center: decor._origin ?? _origin,
        focus: decor._focus ?? _focus,
        shadow: decor._shadow ?? _shadow,
        stroke: decor._stroke ?? _stroke,
        radii: decor._radii ?? _radii,
        strokeSize: decor._strokeSize ?? _strokeSize,
        ink: decor._ink ?? _ink,
      );
    return super.lerp(decor);
  }

  Gradient? get toGradient {
    final decorAp = _DecorApprox(this);
    if (decorAp.paints.isUnitary || decorAp.paints.isEmpty) return null;
    if (data.style == DecorStyle.linear) return LinearGradient(colors: decorAp.paints, begin: decorAp.origin, end: decorAp.focus, stops: decorAp.paints.stops);
    if (data.style == DecorStyle.radial) return RadialGradient(colors: decorAp.paints, focal: decorAp.origin, center: decorAp.focus, stops: decorAp.paints.stops);
    if (data.style == DecorStyle.sweep) return SweepGradient(colors: decorAp.paints, center: decorAp.origin, stops: decorAp.paints.stops);
    return null;
  }

  Decoration get toDecoration {
    final decorAp = _DecorApprox(this);
    return BoxDecoration(
      color: decorAp.paints.isEmpty ? palette.none : decorAp.paints.first,
      borderRadius: decorAp.radii.toBorderRadius,
      gradient: toGradient,
      boxShadow: decorAp.shadow.toBoxShadowDeep,
      border: (decorAp.strokeSize != 0) ? Border.all(color: decorAp.stroke, width: decorAp.strokeSize) : null,
    );
  }

  Widget call() {
    final decorAp = _DecorApprox(this);
    final result = Container(
      child: data.child,
      constraints: data.constraints,
      decoration: toDecoration,
      width: data.width,
      height: data.height,
      clipBehavior: data.clip,
      alignment: data.alignment,
      padding: data.padding,
      margin: data.margin,
    );

    if (_ink == null) return result;
    return (decorAp.ink.$radii(decorAp.radii) & result)();
  }
}

class DecorData extends ConfigData<DecorData> {
  double? width;
  double? height;
  Alignment? alignment;
  EdgeInsets padding;
  EdgeInsets margin;
  BoxConstraints? constraints;
  DecorStyle style;
  Clip clip;

  DecorData({
    Widget? child,
    this.width,
    this.height,
    this.alignment,
    this.padding = EdgeInsets.zero,
    this.margin = EdgeInsets.zero,
    this.constraints,
    this.style = DecorStyle.linear,
    this.clip = Clip.antiAlias,
  }) : super(child);

  @override
  DecorData lerp(decor) {
    if (decor is Size) width = decor.width;
    if (decor is Size) height = decor.height;
    if (decor is Alignment) alignment = decor;
    if (decor is EdgeInsets) padding = decor;
    if (decor is DecorStyle) style = decor;
    if (decor is Clip) clip = decor;

    if (decor is Attr && decor.value is EdgeInsets) {
      if (decor.embedded == #padding) padding = decor.value;
      if (decor.embedded == #margin) margin = decor.value;
    }

    if (decor is DecorData) {
      width = decor.width;
      height = decor.height;
      alignment = decor.alignment;
      padding = decor.padding;
      margin = decor.margin;
      style = decor.style;
      clip = decor.clip;
      constraints = decor.constraints;
    }
    return super.lerp(decor);
  }
}

class DecorTween extends Tween<Decor> {
  DecorTween({Decor? begin, Decor? end}) : super(begin: begin, end: end);

  @override
  Decor transform(double value) => (begin ?? $decor) + ((end ?? $decor) - (begin ?? $decor)) * value;
}

class DecorTransition extends AnimatedWidget {
  final Widget? child;
  DecorTransition({required Animation<Decor> decor, this.child}) : super(listenable: decor);
  Animation<Decor> get decor => listenable as Animation<Decor>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => decor.value.withChild(child)(),
    );
  }
}

enum DecorStyle { linear, radial, sweep }

class DecorPaint extends Color with Subtract<DecorPaint>, Add<DecorPaint> {
  final int value;
  final double stop;

  DecorPaint(this.value, [double stop = 1])
      : stop = stop,
        super(value);

  DecorPaint operator *(double factor) {
    return DecorPaint((Color(value) * factor).value, stop * factor);
  }

  DecorPaint operator -(DecorPaint operand) {
    return DecorPaint((Color(value) - Color(operand.value)).value, stop - operand.stop);
  }

  DecorPaint operator +(DecorPaint operand) {
    return DecorPaint((Color(value) + Color(operand.value)).value, stop + operand.stop);
  }
}

extension DecorColor on Color {
  DecorPaint operator &(double stop) => DecorPaint(value, stop);

  Color operator *(double factor) => Color.fromRGBO((red * factor).round(), (green * factor).round(), (blue * factor).round(), opacity * factor);
  Color operator -(Color color) => Color.fromRGBO(red - color.red, green - color.green, blue - color.blue, opacity - color.opacity);
  Color operator +(Color color) => Color.fromRGBO(red + color.red, green + color.green, blue + color.blue, opacity + color.opacity);
}

extension on List<DecorPaint> {
  double get _totalStop => fold<double>(0, (previousValue, mosaic) => previousValue + mosaic.stop);
  List<double> get stops {
    double former = 0;
    List<double> stops = [];
    for (int i = 0; i < length; i++) {
      DecorPaint mosaic = this[i];
      stops.add((former + mosaic.stop / 2 * i / length) / _totalStop);
      former += mosaic.stop;
    }
    return stops;
  }
}

class _DecorApprox {
  Decor decor;

  _DecorApprox(this.decor);

  List<DecorPaint> get paints => decor._paints ?? [];
  Alignment get origin => decor._origin ?? Alignment.centerLeft;
  Alignment get focus => decor._focus ?? Alignment.centerRight;
  Color get stroke => decor._stroke ?? palette.none;
  double get strokeSize => decor._strokeSize ?? 0;
  Radii get radii => decor._radii ?? $radii;
  Ace get shadow => decor._shadow ?? $ace;
  Ink get ink => decor._ink ?? $ink;
}
