part of nueli.core.config;

/// Tranx shortcut
Tranx get $tranx => Tranx();

/// Tranx ~ [AnimatedWidget] and [Tween] configuration class
class Tranx extends Config<Tranx, TranxData> {
  List<Tween> _tween = [];

  Tranx() : super(TranxData());

  Tranx _config({
    Tween<double>? scale,
    Tween<Offset>? translate,
    Tween<double>? height,
    Tween<double>? width,
    Tween<double>? rotation,
    Tween<double>? opacity,
    Tween<Config>? configTween,
  }) {
    if (scale != null) _tween.add(_Tween.tween(_TweenType.scale, scale));
    if (translate != null) _tween.add(_Tween.tween(_TweenType.translate, translate));
    if (height != null) _tween.add(_Tween.tween(_TweenType.height, height));
    if (width != null) _tween.add(_Tween.tween(_TweenType.width, width));
    if (rotation != null) _tween.add(_Tween.tween(_TweenType.rotation, rotation));
    if (opacity != null) _tween.add(_Tween.tween(_TweenType.opacity, opacity));
    if (configTween != null) _tween.add(configTween);
    return this;
  }

  // variant commands
  Tranx translate(Offset begin, Offset end) => _config(translate: Tween(begin: begin, end: end));
  Tranx scale(double begin, double end, [Alignment alignment = Alignment.center]) => _config(scale: Tween(begin: begin, end: end));
  Tranx size(Size begin, Size end, [Alignment alignment = Alignment.center]) => size_vert(begin.width, end.width).size_horz(begin.height, end.height);
  Tranx rotate(double begin, double end) => _config(rotation: Tween(begin: begin, end: end));
  Tranx fade(double begin, double end) => _config(opacity: Tween(begin: begin, end: end));
  Tranx decor(Decor begin, Decor end) => _config(configTween: DecorTween(begin: begin, end: end));
  Tranx ink(Ink begin, Ink end) => _config(configTween: InkTween(begin: begin, end: end));
  Tranx media(Photo begin, Photo end) => _config(configTween: PhotoTween(begin: begin, end: end));
  Tranx radii(Radii begin, Radii end) => _config(configTween: RadiiTween(begin: begin, end: end));
  Tranx ace(Ace begin, Ace end) => _config(configTween: AceTween(begin: begin, end: end));
  Tranx txt(Txt begin, Txt end) => _config(configTween: TxtTween(begin: begin, end: end));
  Tranx x3D(X3D begin, X3D end) => _config(configTween: X3DTween(begin: begin, end: end));

  // fixed commands
  Tranx withAnimation(Animation<double>? animation) => this & animation;

  // shortcuts
  Tranx size_vert(double begin, double end, [Alignment alignment = Alignment.center]) => _config(width: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names
  Tranx size_horz(double begin, double end, [Alignment alignment = Alignment.center]) => _config(height: Tween(begin: begin, end: end)); // ignore: non_constant_identifier_names

  @override
  @deprecated
  Tranx operator *(double factor) => this;

  @override
  @deprecated
  Tranx operator +(Tranx operand) => this & operand;

  @override
  @deprecated
  Tranx operator -(Tranx operand) => this & operand;

  @override
  Tranx lerp(tranx) {
    if (tranx is Tranx) _tween.addAll(tranx._tween);
    return super.lerp(tranx);
  }

  Widget call() {
    Widget child = data.child ?? SizedBox();
    for (Tween tween in _tween) {
      if (data.animation == null) continue;
      if (tween is _Tween<double>) {
        if (tween.type == _TweenType.scale) child = ScaleTransition(scale: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.height) child = SizeTransition(axis: Axis.vertical, sizeFactor: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.width) child = SizeTransition(axis: Axis.horizontal, sizeFactor: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.opacity) child = FadeTransition(opacity: tween.animate(data.animation!), child: child);
        if (tween.type == _TweenType.rotation) child = RotationTransition(turns: tween.animate(data.animation!), child: child);
      }
      if (tween is _Tween<Offset>) {
        if (tween.type == _TweenType.translate) child = SlideTransition(position: tween.animate(data.animation!), child: child);
      }
      if (tween is Tween<Config>) {
        if (tween is DecorTween) child = DecorTransition(decor: tween.animate(data.animation!), child: child);
        if (tween is InkTween) child = InkTransition(ink: tween.animate(data.animation!), child: child);
        if (tween is PhotoTween) child = PhotoTransition(photo: tween.animate(data.animation!), child: child);
        if (tween is RadiiTween) child = RadiiTransition(radii: tween.animate(data.animation!), child: child);
        if (tween is AceTween) child = AceTransition(ace: tween.animate(data.animation!), child: child);
        if (tween is TxtTween) child = TxtTransition(txt: tween.animate(data.animation!), child: child);
        if (tween is X3DTween) child = X3DTransition(x3D: tween.animate(data.animation!), child: child);
      }
      return child;
    }
    return child;
  }
}

class TranxData extends ConfigData<TranxData> {
  Animation<double>? animation;
  TranxData({Widget? child, this.animation}) : super(child);

  TranxData lerp(data) {
    if (data is Animation<double>) animation = data;
    if (data is TranxData) animation = data.animation;
    return super.lerp(data);
  }
}

enum _TweenType { scale, translate, height, width, rotation, opacity }

class _Tween<T> extends Tween<T> {
  final _TweenType type;
  _Tween(this.type, {T? begin, T? end}) : super(begin: begin, end: end);
  _Tween.tween(this.type, Tween<T> tween) : super(begin: tween.begin, end: tween.end);
}
