library nueli.core.data;

import 'utils/--utils.dart' show AdvanceList, NumCalc;

mixin Subtract<T extends Subtract<T>> {
  T operator -(T operand);
}

mixin Add<T extends Add<T>> {
  T operator +(T operand);
}

mixin Multiply<T extends Multiply<T, Operand>, Operand> {
  T operator *(Operand operand);
}

mixin Divide<T extends Divide<T, Operand>, Operand> {
  T operator /(Operand operand);
}

mixin AND<T extends AND<T, Operand>, Operand> {
  T operator &(Operand operand);
}

mixin OR<T extends OR<T, Operand>, Operand> {
  T operator |(Operand operand);
}

mixin NOT<T extends NOT<T>> {
  T operator -();
}

/// Reference Shortcut
Ref<T> $<T>([T? value]) => Ref<T>.construct(value);

/// <span color = '#99aaff'> Begin Ref ~ </span> [$(Object)]<br>
class Ref<T> {
  static Map<Link, dynamic> _dataMap = {};
  static Link receipt = root;
  final Link link;

  Ref.of(this.link) {
    receipt = link;
    _dataMap.putIfAbsent(link, () => null);
  }

  Ref.construct([T? value]) : link = root / Symbol('${DateTime.now().microsecond}') {
    data = value;
  }

  factory Ref.assign({required Link link, T? value}) {
    return Ref<T>.of(link)..data = value;
  }

  T? get fallback => null;
  T? get data => _dataMap.containsKey(link) ? _dataMap[link] ?? fallback : fallback;

  set data(T? data) => _dataMap[link] = data;

  String toString() => data.toString();
}

Hash<T> $$<T>([Map<Link, T>? value]) => Hash<T>.construct(value);

class Hash<V> extends Ref<Map<Link, V>> {
  Hash.of(Link link) : super.of(link);
  Hash.construct([Map<Link, V>? hash]) : super.construct(hash);
  factory Hash.assign({required Link link, Map<Link, V>? value}) => Hash.of(link)..data = value;

  get fallback => {};

  V? operator [](Link key) => containsKey(key) ? data![key] : null;
  operator []=(Link key, V value) => data![key] = value;

  bool containsValue(V value) => data!.containsValue(value);
  bool containsKey(Link key) => data!.containsKey(key);

  List<MapEntry<Link, V>> get entries => data!.entries.toList();
  String toString() => data.toString();
}

const Link root = Link.create([]);

class Link {
  static const nan = Link.create();

  final List<Symbol> _channel;

  const Link.create([this._channel = const []]);

  Symbol get name => _channel.last;
  Link operator /(Symbol t) => Link.create([..._channel, t]);

  Symbol operator [](int offset) {
    if (_channel.isEmpty) return Symbol.empty;
    assert(offset.modulus + 1 <= _channel.length);
    return (offset > 0) ? _channel[offset] : _channel[_channel.length - 1 + offset];
  }

  operator []=(int offset, Symbol t) {
    if (_channel.isEmpty) return;
    assert(offset.modulus + 1 <= _channel.length);
    offset > 0 ? _channel[offset] = t : _channel[_channel.length - 1 + offset] = t;
  }

  bool startsWith(Link path) => _channel.startsWith(path._channel);
  bool endsWith(Link path) => this._channel.endsWith(path._channel);

  Ref<T> ref<T>(T data) => Ref.assign(link: this, value: data);
  Hash<T> href<T>(Map<Link, T> data) => Hash.assign(link: this, value: data);

  @override
  String toString() => _channel.join('=>');
}
