part of nueli.ui;

/// Screen
abstract class Screen extends StatefulWidget with UI {
  final Ref<bool> _main = $(false);
  final GlobalKey<ScaffoldState> _driver = GlobalKey<ScaffoldState>();

  final Ref<Animation<double>> _animation = $();
  final Ref<Animation<double>> _secondaryAnimation = $();
  final Ref<Animation<double>> _tertiaryAnimation = $();

  final Ref<Duration> _duration = $();
  final Ref<Duration> _secondaryDuration = $();
  final Ref<Duration> _tertiaryDuration = $();

  final Ref<Curve> _curve = $();
  final Ref<Curve> _secondaryCurve = $();
  final Ref<Curve> _tertiaryCurve = $();

  final Ref<Tranx> _tranx = $();
  final Ref<Tranx> _secondaryTranx = $();
  final Ref<Tranx> _tertiaryTranx = $();

  final String name;
  final bool replace;
  final bool closeAllPrev;

  Screen.open(
    BuildContext context, {
    Key? key,
    String? name,
    // tranx
    Tranx? primaryTranx,
    Tranx? secondaryTranx,
    Tranx? tertiaryTranx,
    // curve
    Curve? curve,
    Curve? secondaryCurve,
    Curve? tertiaryCurve,
    // duration
    Duration? duration,
    Duration? secondaryDuration,
    Duration? tertiaryDuration,
    this.replace = false,
    this.closeAllPrev = false,
  })  : name = '${name ?? DateTime.now().millisecond}/@screen',
        super(key: key) {
    // tranx
    _tranx.data = primaryTranx;
    _secondaryTranx.data = secondaryTranx;
    _tertiaryTranx.data = tertiaryTranx;
    // curve
    _curve.data = curve;
    _secondaryCurve.data = secondaryCurve;
    _tertiaryCurve.data = tertiaryCurve;
    // duration
    _duration.data = duration;
    _secondaryDuration.data = secondaryDuration;
    _tertiaryDuration.data = tertiaryDuration;
    Screens.open(context, screen: this, replace: replace, closeAllPrev: closeAllPrev);
  }

  Screen.build({
    Key? key,
    String? name,
    Tranx? primaryTranx,
    Tranx? secondaryTranx,
    Tranx? tertiaryTranx,
    this.replace = false,
    this.closeAllPrev = false,
  })  : name = '${name ?? DateTime.now().millisecond}/@screen',
        super(key: key) {
    _tranx.data = primaryTranx ?? $tranx.translate(Offset(0, 1), Offset.zero).fade(0, 1);
    _secondaryTranx.data = secondaryTranx ?? $tranx;
    _tertiaryTranx.data = tertiaryTranx ?? $tranx.fade(0, 1);
  }

  // Splash
  SplashFactory get splash => SplashFactory.loader();

  // Display
  Decor get displayDecor => $decor.paint_background;
  Alignment get displayAlignment => Alignment.center;
  Decor get displayScrim => $decor.paint_background;

  // Header
  Widget get header => empty;
  Decor get headerDecor => $decor;

  // Body
  Widget get body;
  Widget frame(Widget content) => content;

  // FAB
  Widget get fab => empty;
  FloatingActionButtonLocation get fabLocation => FloatingActionButtonLocation.endFloat;

  // Drawer
  Widget get drawer => empty;
  Decor get drawerDecor => $decor.color_background.$ace(Ace.elevated(8));

  // End Drawer
  Widget get endDrawer => empty;
  Decor get endDrawerDecor => $decor.color_background.$ace(Ace.elevated(8));

  // Tranx
  Tranx get tranx => $tranx.translate(Offset(0, 1), Offset.zero).fade(0, 1);
  Tranx get secondaryTranx => $tranx;
  Tranx get tertiaryTranx => $tranx.fade(0, 1);

  // Curve
  Curve get curve => Curves.easeOutSine;
  Curve get secondaryCurve => curve.flipped;
  Curve get tertiaryCurve => curve;

  // Duration
  Duration get duration => Duration(milliseconds: 800);
  Duration get secondaryDuration => duration;
  Duration get tertiaryDuration => duration;

  // Options
  bool get delayTertiary => true;
  bool get overlayHeader => false;
  bool get useStaticHeader => false;

  @mustCallSuper
  double get headerHeight => !(header == empty) ? headerDecor.data.height ?? 64 : 0;

  @mustCallSuper
  Animation<double> get animation => _animation.data!;

  @mustCallSuper
  Animation<double> get secondaryAnimation => _secondaryAnimation.data!;

  @mustCallSuper
  Animation<double> get tertiaryAnimation => _tertiaryAnimation.data!;

  @mustCallSuper
  void openDrawer() => _driver.currentState!.openDrawer();

  @mustCallSuper
  void openEndDrawer() => _driver.currentState!.openDrawer();

  @mustCallSuper
  close() => Screens.goBack(context);

  @mustCallSuper
  Widget safeRegion({
    Widget? child,
    bool left = true,
    bool top = true,
    bool right = true,
    bool bottom = true,
    SafeArea safeArea = const SafeArea(child: _UITools._empty),
  }) {
    return Padding(
      padding: EdgeInsets.only(
        left: (safeArea.left ? MediaQuery.of(context).viewPadding.left : 0),
        top: (safeArea.top ? MediaQuery.of(context).viewPadding.top : 0) + (top ? headerHeight : 0),
        right: (safeArea.right ? MediaQuery.of(context).viewPadding.right : 0),
        bottom: (safeArea.bottom ? MediaQuery.of(context).viewPadding.bottom : 0),
      ),
    );
  }

  @mustCallSuper
  ScreenRoute get route => ScreenRouteBuilder(screen: this)();

  @mustCallSuper
  open(BuildContext context) {
    return Screens.open(context,
        screen: this, // break
        replace: replace,
        closeAllPrev: closeAllPrev);
  }

  @override
  ScreenState<Screen> createState() => ScreenState();
}

/// Screen state
class ScreenState<T extends Screen> extends ResumableState<T> with UIState<T>, TickerProviderStateMixin {
  final Ref<SplashFactory> _splash = $();
  AnimationController? _controller;

  List<Logic> get logicArray => [];

  Future<bool> onBackPressed() async {
    close();
    return false;
  }

  @mustCallSuper
  void init() {}

  @mustCallSuper
  Future<FutureOr<dynamic>> initAsync() async {
    if (logicArray.isNotEmpty) SplashFactory.invalidate();
    await LogicTree.compile(logicArray);
  }

  @mustCallSuper
  Animation<double> get animation => widget._animation.data ?? AnimationController(vsync: this, value: 1);

  @mustCallSuper
  Animation<double> get secondaryAnimation => widget._secondaryAnimation.data ?? AnimationController(vsync: this, value: 1);

  @mustCallSuper
  Animation<double> get tertiaryAnimation => widget._tertiaryAnimation.data ?? _controller!;

  @override
  setState(VoidCallback fn) => _splash.data!.setState(() => super.setState(fn));

  @mustCallSuper
  double get headerHeight => widget.headerHeight;

  @mustCallSuper
  void openDrawer() => widget._driver.currentState!.openDrawer();

  @mustCallSuper
  void openEndDrawer() => widget._driver.currentState!.openDrawer();

  @mustCallSuper
  close() => Screens.goBack(context);

  @mustCallSuper
  Widget safeRegion({
    Widget? child,
    bool left = true,
    bool top = true,
    bool right = true,
    bool bottom = true,
    SafeArea safeArea = const SafeArea(child: _UITools._empty),
  }) {
    return widget.safeRegion(
      child: child,
      left: left,
      top: top,
      right: right,
      bottom: bottom,
      safeArea: safeArea,
    );
  }

  @mustCallSuper
  void initState() {
    _splash.data = widget.splash;
    _controller = AnimationController(vsync: this, duration: widget.tertiaryDuration);
    init();
    super.initState();
  }

  @mustCallSuper
  Widget build(BuildContext context) {
    super.build(context);

    // Animation
    widget._tertiaryAnimation.data = _controller;

    // Tranx
    widget._tranx.data = widget._tranx.data ?? widget.tranx;
    widget._secondaryTranx.data = widget._secondaryTranx.data ?? widget.secondaryTranx;
    widget._tertiaryTranx.data = widget._tertiaryTranx.data ?? widget.tertiaryTranx;

    // Duration
    widget._duration.data = widget._duration.data ?? widget.duration;
    widget._secondaryDuration.data = widget._secondaryDuration.data ?? widget.secondaryDuration;
    widget._tertiaryDuration.data = widget._tertiaryDuration.data ?? widget.tertiaryDuration;

    // Curve
    widget._curve.data = widget._curve.data ?? widget.curve;
    widget._secondaryCurve.data = widget._secondaryCurve.data ?? widget.secondaryCurve;
    widget._tertiaryCurve.data = widget._tertiaryCurve.data ?? widget.tertiaryCurve;

    WidgetsBinding.instance!.addPostFrameCallback((timeStamp) {
      animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse
          ? !widget.delayTertiary
              ? _controller!.forward()
              : setState(() => _controller!.value = 0)
          : _controller!.forward();
    });
    animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse
        ? !widget.delayTertiary
            ? _controller!.forward()
            : setState(() => _controller!.value = 0)
        : _controller!.forward();
    final plugAnimationIsActive = animation.status == AnimationStatus.forward || animation.status == AnimationStatus.reverse;
    final appBarInstance = Container(
      width: double.infinity,
      child: Stack(fit: StackFit.passthrough, children: [
        (widget.headerDecor & Size(double.infinity, headerHeight + MediaQuery.of(context).viewPadding.top))(),
        Material(color: palette.none, child: SafeArea(child: SizedBox(height: headerHeight, width: double.infinity, child: widget.header))),
      ]),
    );
    final drawer = Drawer(child: widget.drawerDecor.withChild(widget.drawer)());
    final endDrawer = Drawer(child: widget.endDrawerDecor.withChild(widget.endDrawer)());

    final drawerInstance = widget.drawer != empty ? Theme(data: Theme.of(context).copyWith(canvasColor: palette.none), child: drawer) : null;
    final endDrawerInstance = widget.endDrawer != empty ? Theme(data: Theme.of(context).copyWith(canvasColor: palette.none), child: endDrawer) : null;
    final fabInstance = widget.fab != empty ? $tranx.scale(0, 1).fade(0, 1).withAnimation(animation).withChild(widget.fab)() : null;
    final bodyInstance = plugAnimationIsActive ? (widget.delayTertiary ? empty : widget.body) : widget.body;
    final frameInstance = Padding(
      padding: EdgeInsets.only(top: widget.overlayHeader ? 0 : headerHeight),
      child: widget.frame(
        widget.tertiaryTranx.withChild(bodyInstance).withAnimation(CurvedAnimation(
              parent: tertiaryAnimation,
              curve: widget.tertiaryCurve,
              reverseCurve: widget.tertiaryCurve,
            ))(),
      ),
    );

    return Material(
      color: palette.none,
      child: _splash.data!.wrap(
        future: initAsync(),
        child: Stack(children: [
          $tranx.decor($decor, widget.displayScrim).withAnimation(animation)(),
          WillPopScope(
            onWillPop: onBackPressed,
            child: Stack(children: [
              Scaffold(
                key: widget._driver,
                floatingActionButton: fabInstance,
                floatingActionButtonLocation: widget.fabLocation,
                backgroundColor: palette.none,
                drawer: drawerInstance,
                endDrawer: endDrawerInstance,
                body: Stack(children: [
                  widget.tranx.withAnimation(CurvedAnimation(parent: animation, curve: widget.curve, reverseCurve: widget.secondaryCurve)).withChild(
                        widget.secondaryTranx
                            .withAnimation((CurvedAnimation(parent: secondaryAnimation, curve: widget.secondaryCurve, reverseCurve: widget.curve)))
                            .withChild(Stack(children: [frameInstance, widget.useStaticHeader ? empty : appBarInstance]))(),
                      )(),
                  !widget.useStaticHeader ? empty : appBarInstance,
                ]),
              ),
            ]),
          ),
        ]),
      ),
    );
  }

  @override
  void dispose() {
    _controller!.dispose();
    super.dispose();
  }
}

/// ScreenApp
class ScreenApp extends StatelessWidget {
  final Screen entry;
  final bool showDebugSemantics;
  final bool showDebugBanner;
  final Function()? init;

  ScreenApp({
    Key? key,
    required this.entry,
    this.showDebugBanner = false,
    this.showDebugSemantics = false,
    this.init,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    (init ?? () {})();
    return MaterialApp(
      debugShowCheckedModeBanner: showDebugBanner,
      showSemanticsDebugger: showDebugSemantics,
      home: entry.._main.data = true,
      color: palette.background,
      theme: _theme,
    );
  }

  static ThemeData get _theme => ThemeData(
        primaryColor: palette.primary,
        primaryColorDark: palette.primaryDark,
        backgroundColor: palette.background,
        scaffoldBackgroundColor: palette.background,
        colorScheme: ColorScheme.fromSwatch(
          primarySwatch: palette.material as MaterialColor,
          primaryColorDark: palette.primaryDark,
          accentColor: palette.accent,
          backgroundColor: palette.background,
          cardColor: palette.background,
          errorColor: palette.error,
        ).copyWith(
          primary: palette.primary,
          secondary: palette.accent,
          background: palette.background,
          error: palette.error,
          onPrimary: palette.onPrimary,
          onSecondary: palette.onAccent,
          onBackground: palette.onBackground,
          onError: palette.background,
        ),
      );
}

/// Screens
class Screens {
  Screens._();

  static exit() {
    SystemNavigator.pop(animated: true);
  }

  static bool canGoBack(BuildContext context) => Navigator.of(context).canPop();

  static goBack(BuildContext context) {
    if (canGoBack(context)) return Navigator.pop(context);
  }

  static goToPrevModule(BuildContext context) {
    Navigator.of(context).popUntil((route) => (route.settings.name ?? '').endsWith('/@module'));
  }

  static goHome(BuildContext context) {
    Navigator.of(context).popUntil((route) => route.isFirst);
  }

  static goBackTo(BuildContext context, Route route) {
    Navigator.of(context).popUntil((_route) => _route == route);
  }

  static open(BuildContext context, {required Screen screen, bool replace = false, bool closeAllPrev = false}) {
    if (closeAllPrev) {
      Navigator.pushAndRemoveUntil(context, screen.route, (r) => false);
    } else if (replace) {
      Navigator.pushReplacement(context, screen.route);
    } else {
      Navigator.push(context, screen.route);
    }
  }
}

/// Screen route builder
class ScreenRouteBuilder {
  Screen screen;

  ScreenRouteBuilder._(this.screen);
  factory ScreenRouteBuilder({required Screen screen}) => ScreenRouteBuilder._(screen);
  ScreenRoute call() => ScreenRoute(this);
}

/// Screen route
class ScreenRoute extends ModalRoute {
  final ScreenRouteBuilder builder;
  ScreenRoute(this.builder);

  RouteSettings get settings => RouteSettings(name: builder.screen.name);
  Color get barrierColor => palette.none;
  bool get barrierDismissible => false;
  String get barrierLabel => '';

  Widget buildPage(BuildContext context, Animation<double> animation, Animation<double> secondaryAnimation) => builder.screen;

  @override
  Widget buildTransitions(
    BuildContext context,
    Animation<double> animation,
    Animation<double> secondaryAnimation,
    Widget child,
  ) {
    builder.screen._animation.data = animation;
    builder.screen._secondaryAnimation.data = secondaryAnimation;
    return child;
  }

  @override
  bool get maintainState => true;

  @override
  bool get opaque => false;

  @override
  Duration get transitionDuration => builder.screen._duration.data ?? Duration(milliseconds: 350);

  @override
  Duration get reverseTransitionDuration => builder.screen._secondaryDuration.data ?? Duration(milliseconds: 350);
}

/// Splash
class SplashFactory extends Struct {
  static bool _isStartup = true;

  final Ref<bool> _isBusy = $(false);
  final Ref<Widget> _child = $(ui.empty);
  final Ref<Future> _futureAction = $();

  final Widget Function(BuildContext) builder;

  SplashFactory(this.builder);

  factory SplashFactory.loader() {
    return SplashFactory((c) => Material(
        color: palette.background,
        child: Center(
          child: CircularProgressIndicator(
            backgroundColor: palette.accent,
            strokeWidth: 2,
          ),
        )));
  }

  static void invalidate() => _isStartup = true;

  @override
  Widget get body {
    if (_child.data == null) return builder(context);
    return Stack(children: [
      !_isStartup
          ? _child.data!
          : FutureBuilder(
              builder: (c, s) {
                if (s.connectionState == ConnectionState.none) return _child.data ?? empty;
                if (s.connectionState == ConnectionState.waiting) return builder(context);
                _isStartup = false;
                return _child.data ?? empty;
              },
              future: _futureAction.data),
      _isBusy.data! ? Container(color: palette.shade) : empty,
      _isBusy.data! ? builder(context) : empty,
    ]);
  }

  Widget wrap({required Widget child, Future<void>? future}) {
    this._child.data = child;
    this._futureAction.data = future;
    return this;
  }
}

/// Logic
abstract class Logic {
  Symbol get channel;
  Algorithm get initializer;
  Algorithm get reviver => algog;
}

/// Logic Tree
class LogicTree {
  static Map<Symbol, dynamic> _channelBus = {};

  static Future<void> compile(List<Logic> logicArray) async {
    WidgetsFlutterBinding.ensureInitialized();
    for (Logic logic in logicArray) {
      final hasInit = !_channelBus.containsKey(logic.channel);
      _channelBus.putIfAbsent(logic.channel, () => 'init');
      hasInit ? await logic.initializer() : await logic.reviver();
    }
  }
}
